module Clustering (cluster, cross, kMeansCluster)

where
import System.Random (randomIO)
import Maybe ( fromJust )
import Util	
import Prelude hiding (lookup)
import Debug
import Data.Set hiding (delete,intersect,member)
import Data.IntSet hiding (empty,delete)
import Data.Map hiding (map,delete,member)
import Control.Monad.State
import List hiding (insert)

-- first component: map from unique id's to interests
-- second component: map from user id's to usernames (and vice versa)
-- third component: map from user id's to lists of interest id's
-- fourth component: for memoization; maps pairs of usernames onto similarity
-- matches
-- fifth component: also for memoization, maps usernames onto lists of pairs
-- of similarities and usernames, sorted in descending order of similarity
type CM a = StateT (Map Int String, UserMap,
                  Map Int [Int], Map (Int,Int) Double,
                  Map Int [(Int,Double)]) IO a

kMeansCluster :: [String] -> CM [[String]]
-- returns a list of clusters
kMeansCluster users = do
   userIds <- encodeUsers users
   let len = length users
   let k = ((length users) `div` (if (len < 4) then 2 else 4)) + (if (len < 4) then 1 else 0)
   means <- pickMeans k userIds
   let numIterations = 3
   (res::[([Int],[IntSet])]) <- (iterateM ( doKMeans userIds ) (means, [])) numIterations
   lift $ nDebugMsg 1 $ "cluster = " ++ (show res)
   let (_,clusters) = last res
   mapM (decodeUser . Data.IntSet.toList) clusters

pickMeans k users = do
   m <- getInterestsTable
   let sorted = sortBy 
                 (\ u1 u2 -> 
                      compare (length (fromJust (lookupFM m u2))) 
                              (length (fromJust (lookupFM m u1)))) users
   return $ take k sorted
 -- LIFT $ pickRandomMeans k users

pickRandomMeans 0 users = return []
pickRandomMeans k users = do
   (ranNum::Float) <- randomIO
   let random = (floor (ranNum * (fromIntegral (length users))))
   nDebugMsg 1 $ "ranNum = " ++ (show ranNum) ++ " random = " ++ (show random)
   let next = users !! random
   rest <- pickRandomMeans (k-1) (List.delete next users)
   return (users!!random:rest)

doKMeans :: [Int] -> ([Int],[IntSet]) -> CM ([Int],[IntSet])
doKMeans users (means, clusters) = do
   clusters <- assignTo means users
   newMeans <- recomputeMeans clusters
   decodedMeans <- decodeUser means
   decodedNewMeans <- decodeUser newMeans
   lift $ nDebugMsg 0 $ "means = " ++ (show decodedMeans) ++ " newMeans = " ++ (show decodedNewMeans) ++ " cluster = " 
   showClusters clusters
   return (newMeans,clusters)

assignTo :: [Int] -> [Int] -> CM [IntSet]
assignTo means users = do
  let initClusters = List.map Data.IntSet.singleton means
  let clusterNums = zip means ([1..])
  lift $ nDebugMsg 1 $ "clusterNums = " ++ (show clusterNums)
  let usersMinusMeans = deleteFirstsBy (==) users means
  lift $ nDebugMsg 1 $ " means = " ++ (show means) ++ " users = " ++ (show users)
  (bestClusters::[(Int,Double)]) <- mapM (\ user -> 
                        foldM (doIt user clusterNums) (-1,-(1.0)) means) usersMinusMeans
  let result = foldl doIt2 initClusters $ zip usersMinusMeans (fst (unzip bestClusters))
  return result
    where doIt :: Int -> [(Int,Int)] -> (Int,Double) -> Int -> CM (Int,Double)
          doIt user clusterNums x@(bestClusterSoFar,bestScoreSoFar) nextMean = do
                                   sim <- findSimilarity (user,nextMean)
                                   userN <- lookupFMUserId user
                                   meanN <- lookupFMUserId nextMean
                                   lift $ nDebugMsg 2 $ "user = " ++ (show userN) ++ " comparing to mean " ++ (show meanN) ++ " similarity = " ++ (show sim)
                                   (if (sim > bestScoreSoFar) then
                                       return (myFromJust ("doIt: nextMean = " ++ (show nextMean)) (List.lookup nextMean clusterNums),
                                               sim)
                                    else
                                       return x)
          doIt2 :: [IntSet] -> (Int,Int) -> [IntSet]
          doIt2 clustersSoFar (user,bestCluster) =
                       insertInto bestCluster user clustersSoFar

recomputeMeans clusters = mapM recomputeMean clusters

recomputeMean :: IntSet -> CM Int
recomputeMean cluster = do
  let elts = Data.IntSet.elems cluster
  eltsAndScores <- mapM (\ elt -> do
                            sims <- mapM (\ other -> findSimilarity (elt,other)) elts
                            let avgSim = (sum sims)/(fromIntegral (length sims))
                            return (elt,avgSim)) elts
  return $ fst $ (foldl (\ (e1,s1) (e2,s2) -> (if ((max s1 s2) == s1) then (e1,s1) else (e2,s2))) 
                     (head eltsAndScores) 
                     eltsAndScores)
  
cluster :: [String] -> CM [[[String]]]
-- returns a list of lists of clusters (representing diff. levels of the tree)
cluster users = do
   userIds <- encodeUsers users
   lift $ nDebugMsg 2 $ "userIds = " ++ (show userIds)
   let initClusters = List.map Data.IntSet.singleton userIds
   lift $ nDebugMsg 2 $ "initClusters = " ++ (show initClusters)
   let initFm = (addToFM Data.Map.empty 1 initClusters)
   computePairwiseSimilarities userIds
   res <- (clusteringHelper userIds [initFm])
   (decodedRes::[[[String]]]) <- mapM ( decodeUsers . (List.map Data.IntSet.toList)) res
   return decodedRes

-- finite map maps length of cluster onto a list of clusters w/ that length
clusteringHelper :: [Int] -> [Map Int [(IntSet)]] -> CM [[(IntSet)]]
clusteringHelper users clustersSoFar = do
   let lastList = (if (length clustersSoFar > 0) then (last clustersSoFar) else Data.Map.empty)  
   lift $ nDebugMsg 2 $ "clusteringHelper: clustersSoFar = " ++ (show (Data.Map.toList lastList))
   case (lookupFM lastList (length users)) of
      Just _ -> return $ List.map (\x -> (concat (snd (unzip (Data.Map.toList x))))) clustersSoFar
      Nothing -> do 
             lift $ nDebugMsg 2 $ "finding most similar pair..."
             (c1, c2) <- findMostSimilarPair (last clustersSoFar)
             lift $ nDebugMsg 2 $ "finding most similar pair: " ++ (show c1) ++ " , " ++ (show c2)
             let newCluster = c1 `Data.IntSet.union` c2
             lift $ nDebugMsg 2 $ "clustersSoFar = " ++ (show (Data.Map.toList lastList))
             let removed = (removeClusters (last clustersSoFar) [c1, c2]) 
             lift $ nDebugMsg 2 $ "removed = " ++ (show (Data.Map.toList removed))
             let added = addCluster removed newCluster 
             let newClustersSoFar = added
             lift $ nDebugMsg 2 $ "added = " ++ (show (Data.Map.toList added))
             clusteringHelper users (clustersSoFar ++ [newClustersSoFar])

addCluster :: Map Int [(IntSet)] -> IntSet -> Map Int [(IntSet)]
addCluster clusterMap c =
  let len = Data.IntSet.size c in
     case (lookupFM clusterMap len) of
        Just l -> addToFM clusterMap len (c:l) 
        Nothing -> addToFM clusterMap len [c]

removeClusters :: Map Int [(IntSet)] -> [(IntSet)] -> Map Int [(IntSet)]
removeClusters clusterMap cs =
   foldl (\ mapSoFar c -> 
              let len = Data.IntSet.size c in
                case (lookupFM mapSoFar len) of
                  Just l -> addToFM mapSoFar len (delete c l) 
                  Nothing -> mapSoFar) clusterMap cs

  

findMostSimilarPair :: Map Int [(IntSet)] -> CM ((IntSet),(IntSet))
findMostSimilarPair clustersSoFar = do
-- find the smallest cluster so far,
-- only consider clusters of that size
   let totalClusters =  concat (snd (unzip (Data.Map.toList clustersSoFar )))
   let smallestCluster = foldl (\ smallestSoFar next ->
                                   (if ((Data.IntSet.size next) < (Data.IntSet.size smallestSoFar)) then next else smallestSoFar)) (head totalClusters) totalClusters
   let allClusters = (List.filter (\ c -> (Data.IntSet.size c) == (Data.IntSet.size smallestCluster)) totalClusters)
   lift $ nDebugMsg 0 $ " allClusters = " ++ (show allClusters)
   let checkSize = length allClusters /= 1
   foldM ((\ checkSize bestPairSoFar cluster -> do 
             bestOtherCluster <-
                       (foldM (\ bestOtherCluster anotherCluster -> do
                             s1 <- findSimilarityCluster (cluster,anotherCluster)
                             s2 <- findSimilarityCluster (cluster,bestOtherCluster)
                             lift $ nDebugMsg 2 $ "similarity b/w " ++ (show cluster) ++ " and " ++ (show anotherCluster) ++ " is " ++ (show s1)
                             (if( s1 > s2 ) then
                                return anotherCluster
                              else
                                return bestOtherCluster))
                           (head allClusters)
                           allClusters
                           )
-- only merge together 2 clusters of the same size
             (if( (not checkSize) || (Data.IntSet.size cluster == Data.IntSet.size bestOtherCluster) ) then do
                 s1 <- findSimilarityCluster (cluster,bestOtherCluster)
                 s2 <- findSimilarityCluster bestPairSoFar 
                 (if( s1 > s2 ) then
                    return (cluster,bestOtherCluster)
                  else
                    return bestPairSoFar)
              else
                 return bestPairSoFar)) checkSize)
        (head totalClusters, head (tail totalClusters))
        allClusters            

encodeUsers :: [String] -> CM [Int]
encodeUsers users = mapM lookupFMUser users

decodeUsers :: [[Int]] -> CM [[String]]
decodeUsers users = mapM decodeUser users 

decodeUser intList = (mapM lookupFMUserId intList)

lookupFMUser :: String -> CM Int
lookupFMUser user = do
   userMap <- getUserMap
   let res = personToId userMap user
   case res of
       Just i -> return i
       Nothing -> error $ "lookupFMUser: no such user " ++ user

lookupFMUserId :: Int -> CM String                             
lookupFMUserId userId = do
   userMap <- getUserMap
   case (idToPerson userMap userId) of
       Just s -> return s
       Nothing -> error $ "lookupFMUserId: no such user id " ++ (show userId)

findSimilarityCluster :: (IntSet,IntSet) -> CM Double
findSimilarityCluster (c1,c2) = 
   if( c1 == c2 ) then
      return (-1)
   else do
     let lenc2 = Data.IntSet.size c2
     bestScore <- foldM (\ bestSoFar nextElt -> do
                           (similarityList::[(Int,Double)]) <- lookupSimilarityList nextElt
                           -- if ALL scores are worse than best so far, return best so far
                           (if ((snd (head similarityList)) <= bestSoFar) then 
                               return bestSoFar
                            else do
                               bestElt <- (if (lenc2 == 1) then {-# SCC "find_sim" #-} do
                                              let hd = choose c2
                                              sim <- findSimilarity (nextElt, hd)
                                              return $ Just (hd,sim)
                                           else 
                                              return $ {-# SCC "findit" #-} (find (isIn c2) similarityList))
                               case bestElt of
                                   Nothing -> return bestSoFar
                                   Just (_,score) | score > bestSoFar -> return score
                                   _ -> return bestSoFar)) (-1) (Data.IntSet.toList c1)       
--     lift $ nDebugMsg 2 $ "cluster 1 = " ++ (show c1) ++ " cluster 2 = " ++ (show c2) ++ " bestScore = " ++ (show bestScore)                 
     return bestScore
           
isIn :: IntSet -> (Int,a) -> Bool
isIn c2 (elt,_) = elt `member` c2

findSimilarity :: (Int,Int) -> CM Double
findSimilarity (userId1, userId2) =
  (if (userId1 == userId2) then
     return (-1)
   else do
     --lift $ nDebugMsg 2 $ "finding similarity b/w " ++ (show userId1) ++ " and " ++ (show userId2)
     cachedSimilarity <- lookupFMSimilarity userId1 userId2
     --lift $ nDebugMsg 2 $ "cachedSim = " ++ (show cachedSimilarity)
     case cachedSimilarity of
         Just num -> return num
         Nothing -> do
            ints1 <- getInterests userId1
            --lift $ nDebugMsg 2 $ " ints1 = " ++ (show ints1)
            ints2 <- getInterests userId2
            --lift $ nDebugMsg 2 $ " ints2 = " ++ (show ints2)
-- TODO: experiment with different metrics
-- For now just return size of intersection of sets, divided by total size of set
            let numCommon::Double = (fromIntegral (length (ints1 `intersect` ints2))) / (fromIntegral ((length ints1) + (length ints2)))
            setSimilarity userId1 userId2 numCommon
            n1 <- lookupFMUserId userId1
            n2 <- lookupFMUserId userId2
            --lift $ nDebugMsg 2 $ "similarity between " ++ n1 ++ " and " ++ n2 ++ " is " ++ (show numCommon)
            let commInts = ints1 `intersect` ints2
            --decoded <- decodeInterests commInts
            --lift $ nDebugMsg 2 $ "their common interests = " ++ (show decoded)
            setSimilarity userId1 userId2 numCommon
            return numCommon)

computePairwiseSimilarities :: [Int] -> CM ()
computePairwiseSimilarities userIds = do
  mapM_ (\ user -> do
     similarities <- mapM (\ u2 -> do
                             sim <- findSimilarity (user, u2)
                             return (u2,sim)) userIds
     let sortedList = sortBy (\ (_,sim1) (_,sim2) -> compare sim2 sim1) similarities
     setSortedSimilarities user sortedList) userIds 

decodeInterests = mapM decodeInterest

decodeInterest :: Int -> CM String
decodeInterest n = do
  intsMap <- getInterestsMap
  case (lookupFM intsMap n) of
      Just s -> return s
      Nothing -> error $ "decodeInterest: bad interest id " ++ (show n)

getUserMap :: CM UserMap
getUserMap = do
  (_,userMap,_,_,_) <- get
  return userMap

getInterestsMap :: CM (Map Int String)
getInterestsMap = do
  (im,_,_,_,_) <- get
  return im

lookupFMSimilarity :: Int -> Int -> CM (Maybe Double)
lookupFMSimilarity u1 u2 = do
  (simTable::Map Int [(Int,Double)]) <- getSimilarityListTable
  let (mlst::Maybe [(Int,Double)]) = lookupFM simTable u1
  case mlst of 
     Nothing -> return Nothing
     Just lst -> 
        case (find (\ (e,_) -> e == u2) lst) of
           Just (_,d) -> return $ Just d
           Nothing -> return Nothing

setSimilarity :: Int -> Int -> Double -> CM ()
setSimilarity u1 u2 sim = do
   simTable <- getSimilarityTable
   setSimilarityTable (addToFM (addToFM simTable (u1,u2) sim) (u2,u1) sim)

setSortedSimilarities :: Int -> [(Int,Double)] -> CM ()
setSortedSimilarities u sims = do
   simListTable <- getSimilarityListTable
   setSimilarityListTable (addToFM simListTable u sims)

getInterests :: Int -> CM [Int]
getInterests uid = do
   intsTable <- getInterestsTable
   case (lookupFM intsTable uid) of
       Just ints -> return ints
       Nothing -> error $ "getInterests: user id not found in interests table " ++ (show uid)

getSimilarityTable = do
   (_,_,_,st,_) <- get
   return st

setSimilarityTable t = do
   (a,b,c,_,d) <- get
   put (a,b,c,t,d)

getInterestsTable :: CM (Map Int [Int])
getInterestsTable = do
   (_,_,t,_,_) <- get
   return t

getSimilarityListTable :: CM (Map Int [(Int,Double)])
getSimilarityListTable = do
   (_,_,_,_,t) <- get
   return t

setSimilarityListTable :: Map Int [(Int,Double)] -> CM ()
setSimilarityListTable t = do
   (a,b,c,d,_) <- get
   put (a,b,c,d,t)

lookupSimilarityList :: Int -> CM [(Int,Double)]
lookupSimilarityList i = do
   t <- getSimilarityListTable
   case (lookupFM t i) of
     Nothing -> error $ "lookupSimilarityList: bad element " ++ (show i)
     Just x -> return x
 
cross :: IntSet -> IntSet -> Set (Int,Int)   
cross s1 s2 = 
   let l2 = Data.Set.fromList( Data.IntSet.toList s2 ) in
      Data.IntSet.fold (\ x setSoFar ->
                         (Data.Set.map (\ y -> (x,y)) l2) `Data.Set.union` setSoFar) Data.Set.empty s1

choose :: IntSet -> Int
choose = head . Data.IntSet.elems

showClusters clusters = 
   mapM_ (\ c -> do 
      decoded <- decodeUser (Data.IntSet.toList c)
      lift $ nDebugMsg 1 $ show decoded) clusters